unit ArchiversShared;

interface

uses
	Windows, Messages, SysUtils, Types;

procedure SetCursorWait;
procedure ResetCursor;
function WideStrScan(const Str: PWideChar; C: WideChar): PWideChar; register;
function WideExtractFileDir(const FileName: WideString): WideString;
function Max(X, Y: Integer): Integer;
function WideLastDelimiter(const Delimiters, S: WideString): Integer;
function WideCompareFileName(const S1, S2: WideString): Integer;
function WideIncludeTrailingPathDelimiter(const S: WideString): WideString;
function TemporaryPath: WideString;
function UnixUTCToFileTime(UTC: DWORD): FILETIME;

implementation

function Max(X, Y: Integer): Integer;
begin
	if X > Y then Result := X else Result := Y
end;

function TemporaryPath: WideString;
var
	P: array[0..MAX_PATH] of WideChar;
	Len: Integer;
begin
	Len := GetTempPathW(MAX_PATH, P);
	if P[Len - 1] <> '\' then begin P[Len] := '\'; Inc(Len) end;
	SetString(Result, P, Len);
end;

procedure SetCursorWait;
begin
	SetCursor(LoadCursorW(0, PWideChar(IDC_WAIT)));
end;

procedure ResetCursor;
var
	P: TPoint;
	Handle: HWND;
	Code: LRESULT;
	MakeLong: LongRec;
begin
	GetCursorPos(P);
	Handle := WindowFromPoint(P);
	if (Handle <> 0) and
			(GetWindowThreadProcessId(Handle, nil) = GetCurrentThreadId) then
	begin
		Code := SendMessageW(Handle, WM_NCHITTEST, 0, LPARAM(PointToSmallPoint(P)));

		MakeLong.Lo := Code;
		MakeLong.Hi := WM_MOUSEMOVE;

		SendMessageW(Handle, WM_SETCURSOR, Handle, LPARAM(MakeLong));
	end;
end;

function WideStrScan(const Str: PWideChar; C: WideChar): PWideChar; register;
asm
@Loop:
	mov cx, [eax]
	test cx, cx
	jz @NotFound
	cmp cx, dx
	jz @Exit
	inc eax
	inc eax
	jmp @Loop
@NotFound:
	xor eax, eax
@Exit:
end;

function WideLastDelimiter(const Delimiters, S: WideString): Integer;
begin
	Result := Length(S);
	while Result > 0 do
	begin
		if Pos(S[Result], Delimiters) > 0 then Break;
		Dec(Result);
	end;
end;

function WideExtractFileDir(const FileName: WideString): WideString;
var
	I: Integer;
begin
	I := WideLastDelimiter(PathDelim + DriveDelim, FileName);
	if (I > 1) and (FileName[I] = PathDelim) and (((FileName[I - 1] <> PathDelim)
					and (FileName[I-1] <> DriveDelim))) then
		Dec(I);

	Result := Copy(FileName, 1, I);
end;

function WideIncludeTrailingPathDelimiter(const S: WideString): WideString;
begin
	Result := S;
	if (Result = '') or (Result[Length(Result)] <> PathDelim) then
		Result := Result + PathDelim;
end;

function WideCompareFileName(const S1, S2: WideString): Integer;
begin
	Result := CompareStringW(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE,
					Pointer(S1), Length(S1), Pointer(S2), Length(S2)) - 2;
end;

function UnixUTCToFileTime(UTC: DWORD): FILETIME;
const
	Diff = 11644473600;
begin
	Int64(Result) := (UTC + Diff) * Int64(10 * 1000 * 1000);
end;

end.
