unit LZHModule;

interface

uses
	Windows, Messages, SysUtils,
	UnifyArc, Unlha32,
	{$IFNDEF EXTENSION}LObjects, {$ENDIF} EOFileIntf, ArchiversShared;

type
	TFileArray = array of TFile;

	TLZHArchive = class(TInterfacedObject, IFolder, IArchive)
	private
		FPath: TPath;
		FFiles: TFileArray;
		FViews: array of IObserver;
		FValid: Boolean;
		{IObservable}
		procedure Attach(const AView: IObserver);
		procedure Detach(const AView: IObserver);
		function HasClients: Boolean;
		{IFolder}
		function GetPath: TPath;
		function GetItems(AIndex: Integer): PFile;
		function GetCount: Integer;
		function GetValid: Boolean;
		function GetDisplayText: WideString;
		procedure Refresh;
		procedure Move(const NewPath: WideString);
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		procedure StopSurveillance;
		procedure ResumeSurveillance(Wait: Boolean);
		function GetInfoText(AFile: PFile): WideString;
		procedure UnpackMemory(AFile: PFile; var Data: String);
		procedure Unpack(AFile: PFile; const DestDir: WideString);
		function GetPossibilities: TFolderPossibilities;
		function ConcatenatePath(AFile: PFile): TPath;
	protected
		procedure Notify;
		procedure DoRefresh;
	public
		constructor Create(const APath: TPath);
	end;

	TLZHArchiver = class(TInterfacedObject, IExtension, IExtensionInfo, IArchiver)
	public
		{IExtension}
		procedure Initialize;
		procedure Finalize;
		{IExtensionInfo}
		function GetName: WideString;
		procedure Dialog;
		{IArchiver}
		function Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
		function Open(const APath: TPath): IFolder;
		procedure Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
		procedure Unpack(const Archive, DestDir: WideString);
	end;

implementation

function MakeResponseFile(const Files: array of WideString): TFileName;
var
	H: THandle;
	Line: string;
	I: Integer;
begin
	Result := TemporaryPath + '$$$$$$$$.lst';

	H := FileCreate(Result);
	if H = INVALID_HANDLE_VALUE then
		raise EInOutError.CreateFmt('X|Xt@C%s쐬ł܂ł.', [Result]);

	try
		for I := Low(Files) to High(Files) do
		begin
			Line := '"' + string(Files[I]) + '"'#13#10;
			FileWrite(H, PChar(Line)^, Length(Line));
		end;
	finally
		FileClose(H);
	end;
end;

function ExtractArchiveDir(const S: WideString): WideString;
begin
	Result := S;
	Delete(Result, Max(1, WideLastDelimiter('/', Result)), MaxInt);
end;

{ TLZHArchive }

procedure TLZHArchive.Attach(const AView: IObserver);
begin
	SetLength(FViews, Length(FViews) + 1);
	FViews[High(FViews)] := AView;
end;

function TLZHArchive.ConcatenatePath(AFile: PFile): TPath;
var
	S: WideString;
begin
	Result := EmptyPath;

	Result.Directory := FPath.Directory;
	Result.Archive := FPath.Archive;

	with AFile^ do SetString(S, Name, NameLength);

	if Result.Archive <> '' then
		Result.Archive := Result.Archive + '/';

	Result.Archive := Result.Archive + S;
end;

constructor TLZHArchive.Create(const APath: TPath);
begin
	inherited Create;
	FPath := APath;

	SetCursorWait;
	DoRefresh;
	ResetCursor;
end;

procedure TLZHArchive.Detach(const AView: IObserver);
var
	I, J: Integer;
begin
	for I := High(FViews) downto Low(FViews) do
		if FViews[I] = AView then
		begin
			for J := I + 1 to High(FViews) do
				FViews[J - 1] := FViews[J];

			SetLength(FViews, Length(FViews) - 1);
			Break;
		end;
end;

procedure TLZHArchive.DoRefresh;
label
	Next;
var
	Info: TINDIVIDUALINFO;
	Sub: PWideChar;
	I, Count: Integer;
	H: THandle;
	PathInArc: string;
	PathInArcLen: Integer;
	F: PFile;
	PFileName: PChar;
begin
	FValid := False;

	Count := 0;
	SetLength(FFiles, 1024);
	F := @FFiles[0];

	H := UnlhaOpenArchive(0, PChar(string(FPath.Directory)),
					M_REGARDLESS_INIT_FILE or M_NOT_USE_DRIVE_LETTER or
					M_CHECK_FILENAME_ONLY or M_ERROR_MESSAGE_ON or M_NO_BACKGROUND_MODE);
	if H = 0 then	Exit;

	PathInArc := '';
	if FPath.Archive <> '' then PathInArc := FPath.Archive + '/';
	PathInArcLen := Length(PathInArc);

	if UnlhaFindFirst(H, '*', Info) = 0 then
		repeat
			PFileName := Info.szFileName;

			if not CompareMem(PFileName, Pointer(PathInArc), PathInArcLen) then
				Continue;

			Inc(PFileName, PathInArcLen);
			if PFileName^ = #0 then Continue;

			F.NameLength := MultiByteToWideChar(CP_ACP, 0, PFileName, -1, F.Name, High(F.Name)) - 1;
			F.Name[F.NameLength] := #0;

			Sub := WideStrScan(F.Name, '/');
			if Sub = F.Name + F.NameLength - 1 then
			begin
				Dec(F.NameLength);
				F.Name[F.NameLength] := #0;
				Sub := nil;
			end;

			if Sub <> nil then
			begin
				F.NameLength := Sub - F.Name;
				F.Name[F.NameLength] := #0;

				for I := Count - 1 downto 0 do
					if (FFiles[I].NameLength = F.NameLength) and
						CompareMem(@FFiles[I].Name, @F.Name, F.NameLength * SizeOf(WideChar))
					then
						goto Next;

				F.Kind := fkFile;
				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				F.CreationTime.dwLowDateTime := 0;
				F.CreationTime.dwHighDateTime := 0;
				F.LastAccessTime.dwLowDateTime := 0;
				F.LastAccessTime.dwHighDateTime := 0;
				F.LastWriteTime.dwLowDateTime := 0;
				F.LastWriteTime.dwHighDateTime := 0;

				F.Size := 0;
				F.Attributes := faDirectory;

				F.Operatable := [foEnter];
			end
			else
			begin
				F.Kind := fkFile;

				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				UnlhaGetCreateTimeEx(H, F.CreationTime);
				UnlhaGetAccessTimeEx(H, F.LastAccessTime);
				UnlhaGetWriteTimeEx(H, F.LastWriteTime);

				F.Size := Info.dwOriginalSize;
				F.Attributes := UnlhaGetAttribute(H);

				F.Operatable := [];
				if F.Attributes and faDirectory <> 0 then Include(F.Operatable, foEnter);
			end;
			F.Number := High(Word);

			{Add}
			Inc(Count);
			if Count = Length(FFiles) then SetLength(FFiles, Length(FFiles) * 2);
			F := @FFiles[Count];
Next:
		until UnlhaFindNext(H, Info) <> 0;
	UnlhaCloseArchive(H);

	SetLength(FFiles, Count);
	FValid := True;
end;

function TLZHArchive.GetCount: Integer;
begin
	Result := Length(FFiles)
end;

function TLZHArchive.GetDisplayText: WideString;
begin
	Result := FPath.Directory;
	if FPath.Archive <> '' then
		Result := Result + '/' + FPath.Archive
end;

function TLZHArchive.GetInfoText(AFile: PFile): WideString;
begin
	Result := 'in LZH'
end;

function TLZHArchive.GetItems(AIndex: Integer): PFile;
begin
	Result := @FFiles[AIndex]
end;

function TLZHArchive.GetParentPath: TPath;
begin
	Result := FPath;

	if Result.Archive = '' then
		Result.Directory := WideExtractFileDir(Result.Directory)
	else
		Result.Archive := ExtractArchiveDir(Result.Archive)
end;

function TLZHArchive.GetPath: TPath;
begin
	Result := FPath
end;

function TLZHArchive.GetPossibilities: TFolderPossibilities;
begin
	Result := [fpSort, fpHasParent]
end;

function TLZHArchive.GetValid: Boolean;
begin
	Result := FValid
end;

function TLZHArchive.HasClients: Boolean;
begin
	Result := FViews <> nil
end;

function TLZHArchive.IsSub(const AParent: TPath): Boolean;
var
	ParentDir, ParentArc: WideString;
	ParentNameSpace: INameSpace;
begin
	if AParent.ItemIDList = nil then
	begin
		if AParent.Directory = FPath.Directory then
		begin
			ParentArc := AParent.Archive + '/';
			Result := WideCompareFileName(ParentArc, Copy(FPath.Archive + '/', 1, Length(ParentArc))) = 0
		end
		else
		begin
			ParentDir := WideIncludeTrailingPathDelimiter(AParent.Directory);
			Result := WideCompareFileName(ParentDir, Copy(FPath.Directory + '\', 1, Length(ParentDir))) = 0
		end
	end
	else
	begin
		{fXNgbv}CRs[^}
		ParentNameSpace := (GetApplication as IFolderManager).GetFolder(AParent) as INameSpace;
		Result := ParentNameSpace.AtDesktop or ParentNameSpace.AtDrives;
	end;
end;

procedure TLZHArchive.Move(const NewPath: WideString);
begin
	FPath.Directory := NewPath
end;

procedure TLZHArchive.Notify;
var
	I: Integer;
	C: Pointer;
begin
(*
	BeginNotification;
	try
*)
		{EĂvȂ悤Ɂc}
		I := Low(FViews);
		while I <= High(FViews) do
		begin
			C := Pointer(FViews[I]);
			FViews[I].Observe;
			if C = Pointer(FViews[I]) then
				Inc(I);
		end;
(*
	finally
		EndNotification
	end
*)
end;

procedure TLZHArchive.Refresh;
var
	OldFiles: TFileArray;
begin
	if FValid then
	begin
		SetCursorWait;

		OldFiles := FFiles; {View ܂QƂĂ\邽߉Ȃ}
		try
			DoRefresh;
		except
			FValid := False;
		end;
		ResetCursor;

		Notify; {ʒm}
	end;

	OldFiles := nil; {sage...}
end;

procedure TLZHArchive.ResumeSurveillance;
begin
end;

procedure TLZHArchive.StopSurveillance;
begin
end;

procedure TLZHArchive.Unpack(AFile: PFile; const DestDir: WideString);
var
	DestPath, F, Full: WideString;
	Cmd: string;
	Buffer: array[0..4095] of Char;
begin
	DestPath := WideIncludeTrailingPathDelimiter(DestDir);

	with AFile^ do SetString(F, Name, NameLength);
	if FPath.Archive <> '' then F := FPath.Archive + '/' + F;

	Full := DestPath + F;

  {-y: ㏑}

	Cmd := 'e -y "' + FPath.Directory + '" "' + DestPath + '" "' + F + '"';
	if Unlha(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise EUnlhaError.Create(Buffer);
end;

procedure TLZHArchive.UnpackMemory(AFile: PFile; var Data: String);
var
	Size: DWORD;
	Cmd: string;
	F: WideString;
	E: Integer;
begin
	Size := AFile^.Size;
	SetLength(Data, Size);

	with AFile^ do SetString(F, Name, NameLength);
	if FPath.Archive <> '' then F := FPath.Archive + '/' + F;

	Cmd := '"' + FPath.Directory + '" "' + F + '"';

	E := UnlhaExtractMem(0, PChar(Cmd), Pointer(Data), Size, nil, nil, nil);
	if E <> 0 then
		raise EUnlhaError.Create('UnlhaExtractMem: ' + IntToStr(E));
end;

{ TLZHArchiver }

procedure TLZHArchiver.Dialog;
var
	Buffer: array[0..1023] of Char;
begin
	try
		UnlhaConfigDialog(0, Buffer, UNPACK_CONFIG_MODE)
	except
		MessageBoxW(0, 'Unlha32.dll܂', 'LZH', MB_ICONERROR)
	end
end;

function TLZHArchiver.Enumerate(Proc: CEachArchiveKind;	Context: Pointer): WideString;
begin
	if Proc('.lzh', Context) then
		Result := '.lzh'
	else
		Result := ''
end;

procedure TLZHArchiver.Finalize;
begin
end;

function TLZHArchiver.GetName: WideString;
begin
	Result := 'Unlha32.dll'
end;

procedure TLZHArchiver.Initialize;
begin
end;

function TLZHArchiver.Open(const APath: TPath): IFolder;
begin
	if not UnlhaEnabled then
		raise EUnlhaError.CreateFmt(SFaultLoad, [Unlha32Dll]);

	Result := TLZHArchive.Create(APath)
end;

procedure TLZHArchiver.Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
var
	Path, Cmd, ResponseFile: String;
	Buffer: array[0..4095] of Char;
begin
	try
		Path := IncludeTrailingPathDelimiter(SourceDir);

		Cmd := 'u -d1 "' + Archive +'" "' + Path + '"';
		if Length(Sources) = 1 then
			Cmd := Cmd + ' "' + Sources[0] +'"'
		else
		begin
			ResponseFile := MakeResponseFile(Sources);
			Cmd := Cmd + ' @"' + ResponseFile +'"';
		end;

		if Unlha(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
			raise EUnlhaError.Create(Buffer);
	finally
		if ResponseFile <> '' then Windows.DeleteFile(PChar(ResponseFile));
	end
end;

procedure TLZHArchiver.Unpack(const Archive, DestDir: WideString);
var
	Cmd: String;
	Buffer: array[0..4095] of Char;
begin
	Cmd := 'x -a1 -x1 -jsp3 "' + Archive + '" "' + IncludeTrailingPathDelimiter(DestDir) + '"';

	if Unlha(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise EUnlhaError.Create(Buffer);
end;

procedure Register;
begin
	(GetApplication as IExtensionManager).RegisterExtension(TLZHArchiver.Create)
end;

initialization

	if GetInterfaceVersion = EOFileIntf.InterfaceVersion then
		Register

end.
